import os
import platform
import pytest
import functools

import helpers
import pqclean

__tracebackhide__ = True

if platform.machine() == "ppc":
    pytest.skip("Skipping this test on PowerPC to save cycles.", allow_module_level=True)


__astyle_version_result = None

def _get_astyle_version() -> str:
    """Get the AStyle version number"""
    # functools.lru_cache doesn't work because we want to print to stdout each time.
    global __astyle_version_result

    if __astyle_version_result is None:
        __astyle_version_result = helpers.run_subprocess(['astyle', '--version'])
    else:
        print(__astyle_version_result)
    return __astyle_version_result


helpers.ensure_available('astyle')

# Check AStyle version
def version_check():
    result = _get_astyle_version()
    if ("Artistic Style Version 3.4.8" in result or
        "Artistic Style Version 3.4.9" in result or
        "Artistic Style Version 3.4.10" in result):
        return False
    return "Artistic Style Version 3.4" in result

if not version_check() and "CI" not in os.environ:
    pytest.skip("Incompatible AStyle version (need 3.4.x) (not 3.4.{8-10})", allow_module_level=True)

if "CI" in os.environ and platform.system() in { 'Windows', 'Darwin' }:
    pytest.skip("Skipping AStyle on Windows and MacOS in CI", allow_module_level=True)

@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_format(implementation: pqclean.Implementation):
    cfiles = implementation.cfiles()
    hfiles = implementation.hfiles()
    # Triggers a print
    _get_astyle_version()
    result = helpers.run_subprocess(
        ['astyle',
         '--dry-run',
         '--options=../.astylerc',
         *cfiles,
         *hfiles])
    assert 'Formatted' not in result


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



import pytest

import helpers
from pathlib import Path
import pqclean


@pytest.mark.parametrize(
    'implementation',
    pqclean.Scheme.all_implementations(),
    ids=str,
)
@helpers.filtered_test
def test_preprocessor(implementation: pqclean.Implementation):
    cfiles = implementation.cfiles()
    hfiles = implementation.hfiles()
    errors = []
    for file in hfiles + cfiles:
        code_has_begun = False
        in_comment = True
        with open(file) as f:
            for i, line in enumerate(f):
                line = line.strip()
                if file in hfiles and not code_has_begun and line.startswith('#ifndef'):
                    continue
                if line == "":
                    continue
                if line.startswith("/*") and not code_has_begun:
                    in_comment = True
                    continue
                if in_comment and "*/" in line:
                    in_comment = False
                    continue
                if not code_has_begun and not in_comment and not line.startswith("#ifndef"):
                    code_has_begun = True
                if (line.startswith('#if') and 
                    not Path(file).match("crypto_sign/dilithium[235]/aarch64/fips202x2.c") and
                    not Path(file).match("crypto_kem/kyber*/aarch64/fips202x2.c") and
                    not Path(file).match("crypto_sign/falcon-*/aarch64/fips202x2.c") and
                    not Path(file).match("crypto_sign/falcon-*/avx2/sign.c")):
                    errors.append("\n at {}:{}".format(file, i+1))
    if errors:
        raise AssertionError(
            "Prohibited use of preprocessor conditional" + "".join(errors)
        )


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)



"""
Checks that every .c and .h file in an implementation is present as a
dependency of that scheme's Makefile.
"""

import datetime
import glob
import os

import pytest

import helpers
import pqclean


@pytest.mark.parametrize(
    'implementation,test_dir,impl_path, init, destr',
    [(impl,
      *helpers.isolate_test_files(impl.path(), 'test_makefile_deps_'))
     for impl in pqclean.Scheme.all_supported_implementations()],
    ids=[str(impl) for impl in pqclean.Scheme.all_supported_implementations()],
)
@helpers.filtered_test
def test_makefile_dependencies(implementation, impl_path, test_dir,
                               init, destr):
    init()
    # initial build - want to have *all* files in place at beginning
    helpers.make('clean', working_dir=impl_path)
    helpers.make(working_dir=impl_path)
    # test case for each candidate file
    cfiles = glob.glob(os.path.join(impl_path, '*.c'))
    sfiles = glob.glob(os.path.join(impl_path, '*.[sS]'))
    incfiles = glob.glob(os.path.join(impl_path, '*.inc'))
    hfiles = glob.glob(os.path.join(impl_path, '*.h'))
    try:
        for file in (cfiles + hfiles + sfiles + incfiles):
            check_makefile_dependencies(implementation, impl_path, file)
    except Exception as e:
        print("Affected file: {}".format(file))
        raise e
    finally:
        destr()


def touch(time, *files):
    for path in files:
        times = (time.timestamp(), time.timestamp())
        os.utime(path, times)


def make_check(path, expect_error=False):
    makeflag = '-q' if os.name != 'nt' else '/Q'
    expected_returncode = 0
    if expect_error:
        expected_returncode = 1 if os.name != 'nt' else 255
    helpers.make(makeflag, working_dir=path,
                 expected_returncode=expected_returncode)


def check_makefile_dependencies(implementation, impl_path, file):
    cfiles = glob.glob(os.path.join(impl_path, '*.c'))
    sfiles = glob.glob(os.path.join(impl_path, '*.[sS]'))
    hfiles = glob.glob(os.path.join(impl_path, '*.h'))
    incfiles = glob.glob(os.path.join(impl_path, '*.inc'))
    o_ext = '*.o' if os.name != 'nt' else '*.obj'
    ofiles = glob.glob(os.path.join(impl_path, o_ext))

    # handle dependency files: these also need to be set correctly
    commondir = os.path.join(impl_path, '..', '..', '..', 'common')
    cfiles += glob.glob(os.path.join(commondir, '*.c'))
    cfiles += glob.glob(os.path.join(commondir, '**', '*.c'))
    hfiles += glob.glob(os.path.join(commondir, '*.h'))
    hfiles += glob.glob(os.path.join(commondir, '**', '*.h'))
    incfiles.append(os.path.join(commondir, 'keccak4x', 'KeccakP-1600-unrolling.macros'))
    ofiles += glob.glob(os.path.join(commondir, o_ext))
    ofiles += glob.glob(os.path.join(commondir, '**', o_ext))

    libfile = os.path.join(impl_path, implementation.libname())

    # modification time-based calculations is tricky on a sub-second basis
    # so we reset all the modification times to a known and "sensible" order
    now = datetime.datetime.now() - datetime.timedelta(seconds=10)
    ago15 = now - datetime.timedelta(minutes=15)
    ago10 = now - datetime.timedelta(minutes=10)
    ago5 = now - datetime.timedelta(minutes=5)

    touch(ago15, *cfiles, *sfiles, *hfiles, *incfiles)
    touch(ago10, *ofiles)
    touch(ago5, libfile)

    # Sanity check: the scheme is up to date
    make_check(impl_path)

    # touch the candidate .c / .h file
    touch(now, file)

    # check if it needs to be rebuilt using make -q
    make_check(impl_path, expect_error=True)


if __name__ == '__main__':
    import sys
    pytest.main(sys.argv)
